/*
 * Copyright (c) 1997 Regents of the University of Michigan.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that this notice is preserved and that due credit is given
 * to the University of Michigan at Ann Arbor. The name of the University
 * may not be used to endorse or promote products derived from this
 * software without specific prior written permission. This software
 * is provided ``as is'' without express or implied warranty.
 *
 * krb4_auth.so: Kerberos v4 authentication,
 *               AFS PTS group authorization.
 *               Modifications to work with our
 *               Kerberos browser plug-in.
 *
 * See http://www-personal.umich.edu/~dugsong/krb4nsapi for more info.
 *
 * Copyright (c) 1997 Dug Song <dugsong@UMICH.EDU>
 * All rights reserved, all wrongs reversed.
 * Disclaimer: i claim dis
 */

#include "base/pblock.h"
#include "base/session.h"
#include "base/util.h"
#include "frame/req.h"
#include "frame/http.h"
#include "frame/log.h"

/* Standard Kerberos includes. */
#include <krb.h>

#ifdef KRB_DEBUG
#define D(x,y) log_error(LOG_INFORM, x, sn, rq, y)
#else
#define D(x,y)
#endif /* KRB_DEBUG */

#define PLUGIN
#define NETSCAPE_ONLY

#ifdef XP_WIN32
#define NSAPI_PUBLIC __declspec(dllexport)
#else /* !XP_WIN32 */
#define NSAPI_PUBLIC
#endif /* !XP_WIN32 */

#define EMBED_FMT "<EMBED %s PLUGINSPAGE=\"%s\" PRINCIPAL=\"%s\" URL=\"%s\"%s></EMBED>\n"
#ifndef NETSCAPE_ONLY
#define EMBED_MIME "SRC=\"/dummy.krb4\""
#else
#define EMBED_MIME "TYPE=\"application/x-kerberos4-mime\""
#endif
#define MAX_POST_DATA 2048
#define KRB_AUTH_REFRESH 5
#define PLUGIN_PAGE "http://www.umich.edu/~dugsong/kplugin/"
#define PLUGIN_DENIAL_MSG	"<H1>Authorization Required</H1>\n<P>\n \
This server could not verify that you are authorized to access the \n \
document you requested. Either you supplied the wrong credentials, or \n \
your browser doesn't understand how to supply the credentials required. \n \
Please refer to <a href=\"http://www.umich.edu/~dugsong/kplugin/\"> \n \
http://www.umich.edu/~dugsong/kplugin/</a> for information on \n \
how to enable your browser to perform Kerberos authentication.\n<P>\n"
#define PLUGIN_NOSCRIPT_MSG	"<NOSCRIPT>\n<h1>Kerberos v4 authentication \
error</h1>\n<p>\nYou need to have a Javascript-capable browser to use the \
Kerberos plug-in. If you are using Navigator 2.0 or later and you see this \
message, you need to enable JavaScript.<p>\n Choose <b>Options -> Network \
Preferences</b>, choose the <b>Languages</b> tab and click <b>Enable \
JavaScript</b>.\n</NOSCRIPT>\n"

/* Convert radix-encoded ticket to ktext struct. */
#define GETLONG(l, cp) { \
register unsigned char *t_cp = (unsigned char*)(cp); \
(l) = (((unsigned int)t_cp[0]) << 24) \
    | (((unsigned int)t_cp[1]) << 16) \
    | (((unsigned int)t_cp[2]) << 8) \
    | (((unsigned int)t_cp[3])) \
    ; \
}

int radix_to_tkt(char *buf, KTEXT auth)
{
  int len, version;
  char *t, tmp[1024];

  len = uudecode(tmp, buf, 1024);
  t = tmp;
  version = *t; t++; len--;
  GETLONG(auth->length, t); t += 4; len -= 4;
  if (auth->length < 0 || auth->length > len ||
      auth->length > sizeof(auth->dat)) return 1;
  memcpy(&auth->dat, t, auth->length);
  return 0;
}

/* Verify HMAC-MD5 of URL(+QUERY_STRING, if any). */
int krb4_check_hmac_md5(Session *sn, Request *rq)
{
  char digest[33], uri[MAXPATHLEN];
  char *hmac = pblock_findval("auth-checksum", rq->vars);
  unsigned char *session =
    (unsigned char *)pblock_findval("auth-session", rq->vars);
  
  strncpy(uri, pblock_findval("uri", rq->reqpb), sizeof(uri)-1);
  hmac_md5(uri, strlen(uri), session, DES_KEY_SZ, digest);
#ifdef PLUGIN
  if (strncmp(hmac, digest, 32) != 0) {
    if (uri[strlen(uri)] != '/') {
      *(strrchr(uri, '/')+1) = '\0';
    }
    hmac_md5(uri, strlen(uri), session, DES_KEY_SZ, digest);
  }
#endif /* PLUGIN */
  
  return strncmp(hmac, digest, 32);
}

int krb4_auth_required(Session *sn, Request *rq, char *msg)
{
  pb_param *pp;
  if ((pp = pblock_remove("krb4err", rq->vars)) != NULL) param_free(pp);
  pblock_nvinsert("krb4err", msg, rq->vars);
  D("krb4err", msg);
  return REQ_NOACTION;
}

#ifdef PLUGIN
/* Generate refreshed cookie for client. */
void krb4_make_cookie (Session *sn, Request *rq, const char *ticket,
		       const char *checksum, const time_t expires)
{
  char cookie[1024], gmtimestr[32];
  time_t when = expires + KRB_AUTH_REFRESH;
  struct tm *tms = gmtime(&when);

  util_strftime(gmtimestr, "%a, %d %b %Y %T GMT", tms);
  
  util_snprintf(cookie, 1024, "KerberosV4=\"ticket=\"%s\"checksum=\"%s\""
		"expires=\"%d\"\"; expires=%s; path=/",
		ticket, checksum, when, gmtimestr);
  D("returning cookie", cookie);
  pblock_nvinsert("Set-Cookie", cookie, rq->srvhdrs);
}
#endif /* PLUGIN */

NSAPI_PUBLIC int krb4_auth_check(pblock *pb, Session *sn, Request *rq)
{
  KTEXT_ST ktext;
  AUTH_DAT adat = { 0 };
  char *dest_uri, *auth_line, *tmp, *envline;
  char *srvtab, *principal, *ticket, *checksum, *expires;
  char service[ANAME_SZ], instance[INST_SZ], realm[REALM_SZ];
  char *require_ssl;
  int rc;
  pb_param *session_pbp;

  /* Get the destination URI the user is trying to retrieve */
  if ((dest_uri = pblock_findval("uri", rq->reqpb)) == NULL) {
    protocol_status(sn, rq, PROTOCOL_BAD_REQUEST, NULL);
    return REQ_ABORTED;
  }
  /* Check if we require SSL. */
  if ((require_ssl = pblock_findval("requires_ssl", pb)) &&
      strncasecmp(require_ssl, "on", 2) == 0 &&
      !pblock_findval("cipher", sn->client)) {
    pblock_nvinsert("auth-user", "", rq->vars); /* Client should see this. */
    return krb4_auth_required(sn, rq, "authentication refused for non-SSL "
			      "encrypted connection");
  }
  /* Check for auth info in Authorization: or Cookie: headers. */
  rc = request_header("authorization", &auth_line, sn, rq);
#ifdef PLUGIN
  if (rc == REQ_ABORTED || auth_line == NULL)
    rc = request_header("cookie", &auth_line, sn, rq);
#endif /* PLUGIN */

  /* Make sure client specified right authentication method. */
  if (rc == REQ_ABORTED || auth_line == NULL ||
      strncasecmp(auth_line, "KerberosV4", 10) != 0) {
    return krb4_auth_required(sn, rq, "wrong authentication type!");
  }
  /* Get base64-encoded ticket and HMAC from auth line. */
  tmp = STRDUP(auth_line);
  strtok(strstr(tmp, "ticket=\""), "\"");
  ticket = strtok(NULL, "\"");
  strtok(strstr(tmp, "checksum=\""), "\"");
  checksum = strtok(NULL, "\"");
  strtok(strstr(tmp, "expires=\""), "\"");
  expires = strtok(NULL, "\"");
  D("auth_line", auth_line);
  
  /* Make sure we got ticket and checksum. */
  if (!ticket || !checksum) {
    return krb4_auth_required(sn, rq, "no ticket or checksum!");
  }
  /* Decode base-64 encoded ticket. */
  if (radix_to_tkt(ticket, &ktext) != 0) {
    return krb4_auth_required(sn, rq, "uudecode of ticket failed!");
  }
  /* Get our srvtab file location and Kerberos principal. */
  *service = '\0'; *instance = '\0'; *realm = '\0';
  if (!(srvtab = pblock_findval("srvtab", pb)) ||
      !(principal = pblock_findval("principal", pb)) ||
      kname_parse(service, instance, realm, principal) != KSUCCESS) {
    protocol_status(sn, rq, PROTOCOL_SERVER_ERROR, "Kerberos configuration");
    return REQ_ABORTED;
  }
  /* Verify authenticator. */
  if ((rc = krb_rd_req(&ktext, service, instance, 0L, &adat, srvtab))){
    pblock_nvinsert("auth-user", "", rq->vars); /* Client should see this. */
    return krb4_auth_required(sn, rq, krb_err_txt[rc]);
  }
#ifdef PLUGIN
  /* For plug-in POSTs, return cookie for images in this protection space. */
  if (expires) krb4_make_cookie(sn, rq, ticket, checksum, atol(expires));
#endif /* PLUGIN */
  /* Save cookie and auth results for later. */
  envline = (char *)MALLOC(12 + strlen(auth_line) + 1);
  util_sprintf(envline, "HTTP_COOKIE=%s", auth_line);
  putenv(envline);

  putenv("AUTH_TYPE=KerberosV4");
  pblock_nvinsert("auth-type", "KerberosV4", rq->vars);

  envline = (char *)REALLOC(envline, 10 + strlen(adat.pname) + 1);
  util_sprintf(envline, "AUTH_USER=%s", adat.pname);
  putenv(envline); 
  pblock_nvinsert("auth-user", adat.pname, rq->vars);

  session_pbp = (pb_param *)MALLOC(sizeof(pb_param));
  session_pbp->name = "auth-session";
  session_pbp->value = (char *)MALLOC(sizeof(C_Block));
  memcpy(session_pbp->value, adat.session, DES_KEY_SZ);
  pblock_pinsert(session_pbp, rq->vars);

  pblock_nvinsert("auth-checksum", checksum, rq->vars);
  
  return REQ_PROCEED;
}

NSAPI_PUBLIC int krb4_path_check(pblock *pb, Session *sn, Request *rq)
{
  char *user = pblock_findval("auth-user", rq->vars);
  char *type = pblock_findval("auth-type", rq->vars);
  char *path = pblock_findval("path", rq->vars);
  char *enforce_hmac = pblock_findval("enforce_hmac", pb);
  char *command, reason[64];

  /* Check to see if URL even exists. */
  if (!path || !request_stat_path(path, rq)) {
    protocol_status(sn, rq, PROTOCOL_NOT_FOUND, NULL);
    return REQ_ABORTED;
  }
  /* Check for failure in AuthTrans step, fall thru to custom handler. */
  if (!user || ! type || pblock_findval("krb4err", rq->vars) ||
      strncmp(type, "KerberosV4", 10) != 0) {
    return REQ_NOACTION;
  }
  /* Verify HMAC. */
  if (enforce_hmac && strncmp(enforce_hmac, "on", 2) == 0) {
    if (krb4_check_hmac_md5(sn, rq) != 0) {
      pblock_remove("auth-user", rq->vars);
      return krb4_auth_required(sn, rq, "invalid HMAC-MD5!");
    }
  }
  /* Check .htaccess/AFS directory ACLs. This is a super-kludge,
     feel free to replace this stanza with a call to a real authorization
     scheme. We can't include AFS routines in this module because of the
     problems in trying to create a shared object. */
  
  command = (char *)MALLOC(strlen(PATH_ACLTEST)+strlen(user)+strlen(path)+3);
  util_sprintf(command, "%s %s %s", PATH_ACLTEST, user, path);
  if (system(command) == 0) return REQ_NOACTION;
  
  /* Set response code to forbidden and return an error. */
  util_snprintf(reason, sizeof(reason),
		"user %s is not allowed to access this URL", user);
  return krb4_auth_required(sn, rq, reason);
}

void reset_content_type(Request *rq)
{
  pb_param *pp;
  if ((pp = pblock_remove("content-type", rq->srvhdrs)) != NULL)
    param_free(pp);
  pblock_nvinsert("content-type", "text/html", rq->srvhdrs);
}

char *get_post_data(Session *sn, Request *rq)
{
  char *pdata = "";
  char *clen = pblock_findval("content-length", rq->headers);
  int i, len = atoi(clen ? clen : "0");
  
  if (len && len < MAX_POST_DATA &&
      (pdata = (char *)MALLOC(len + 3)) != NULL) {
    for (i = 0; i < len; i++)
      pdata[i] = netbuf_getc(sn->inbuf);
    pdata[i] = '\0';
  }
  return pdata;
}

NSAPI_PUBLIC int krb4_service_reply(pblock *pb, Session *sn, Request *rq)
{
  char *user, *krb4err, *user_agent;
  char *principal, *url, *method, *errdoc;
  char challenge[BUFSIZ], post[MAX_POST_DATA];
  char *allow_client_caching = pblock_findval("allow_client_caching", pb);
  struct stat *st;
  SYS_FILE fd;
  filebuf *fb;
  
  /* Tell the client not to cache our response. */
  if (!allow_client_caching || strncmp(allow_client_caching, "on", 2) != 0) {
    pblock_nvinsert("pragma", "no-cache", rq->srvhdrs);
    pblock_nvinsert("cache-control", "no-cache", rq->srvhdrs);
  }
  /* If no Kerberos authentication errors, grant access. */
  if ((krb4err = pblock_findval("krb4err", rq->vars)) == NULL) {
    D("krb4_service_reply", "allowing access!");
    return REQ_NOACTION;
  }
  user = pblock_findval("auth-user", rq->vars);
  principal = pblock_findval("principal", pb);
  url = http_uri2url_dynamic(pblock_findval("uri", rq->reqpb), "", sn, rq);
  method = pblock_findval("method", rq->reqpb);
  
  /* Return authorization required code and auth challenge. */
#ifdef PLUGIN
  if (request_header("user-agent", &user_agent, sn, rq) != REQ_ABORTED &&
      util_is_mozilla(user_agent, "3", "0")) {
    reset_content_type(rq);

    /* Generate challenge. */
    if (method && strncmp(method, "POST", 4) == 0)
      util_snprintf(post, MAX_POST_DATA, " POST=\"%s\"",
		    get_post_data(sn, rq));
    else strcpy(post, "");
    util_snprintf(challenge, BUFSIZ, EMBED_FMT, EMBED_MIME, PLUGIN_PAGE,
		  principal, url, post);
    protocol_status(sn, rq, PROTOCOL_BAD_REQUEST, "Kerberos auth required");
    if (protocol_start_response(sn, rq) != REQ_NOACTION) {
      if (!user)(void)net_write(sn->csd, "<NOEMBED>\n", 10);
      if ((errdoc = pblock_findval("error_document", pb)) != NULL &&
	  (st = request_stat_path(errdoc, rq)) != NULL &&
	  (fd = system_fopenRO(errdoc)) != IO_ERROR &&
	  (fb = filebuf_open_nostat(fd, FILE_BUFFERSIZE, st)) != NULL) {
	filebuf_buf2sd(fb, sn->csd);
	filebuf_close(fb);
	system_fclose(fd);
      }
      else net_write(sn->csd, PLUGIN_DENIAL_MSG, strlen(PLUGIN_DENIAL_MSG));
      net_write(sn->csd, krb4err, strlen(krb4err));
      if (!user) {
	(void)net_write(sn->csd, "</NOEMBED>\n", 11);
	(void)net_write(sn->csd, challenge, strlen(challenge));
      }
      net_write(sn->csd, PLUGIN_NOSCRIPT_MSG, strlen(PLUGIN_NOSCRIPT_MSG));
    }
    return REQ_PROCEED;
  }
#endif /* PLUGIN */
  if (!(principal = pblock_findval("principal", pb))) {
    protocol_status(sn, rq, PROTOCOL_SERVER_ERROR, "Kerberos error");
    return REQ_ABORTED;
  }
  util_snprintf(challenge, BUFSIZ, "KerberosV4 realm=\"%s\"", principal);
  pblock_nvinsert("WWW-authenticate", challenge, rq->srvhdrs);
  protocol_status(sn, rq, PROTOCOL_UNAUTHORIZED, NULL);
  return REQ_ABORTED;
}

/* 5000, yo. */
